function [R,t,v,a,Thistory] = ecmpr(X,Y,maxNumIter)
% ECMPR Expectation Conditional Maximization - Point Registration. 
%    Maximum Likelihood estimation of trasformation parameters for the 
%    rigid (up to an isometric transform) registration of point sets X,Y.
%
%    [R,t] = ecmpr(X,Y)  Given an [3 x n] matrix X and a [3 x m] matrix Y,
%    computes a [3 x 3] rotation matrix R and [3 x 1] translation vector t 
%    in the sense that Y ~= R*X+t*ones(1,n).
%
%    [R,t,v] = ecmpr(X,Y) Returns also an [n x 1] vector with elements corr
%    esponding to the final variances (sigma_i^2) of the isotropic clusters
%    of the GMM of X ~ N(x_i,v*n).
%
%    [R,t,v,a] = ecmpr(X,Y) Returns also an [m x n] matrix with the final 
%    posterior probabilities of the assignment variable Z_i. In specfic 
%    a(j,i) is the probablity that a data point Y(:,j) is assigned to the 
%    GMM component X(:,i) or in other words a(j,i)= p( Z_i = j | Y(:,j) )
%
%    [R,t,v,a,Thistory] = ecmpr(X,Y) Returns also a [maxNumIter x 2] cell 
%    array containing the transformation parameters computed at each
%    iteration. Note that Thistory{i,1} contains R and Thistory{i,2} 
%    contains t computed at iteration 'i'.
%    Note that Rhistory is allocated only if it is requested. 
%
%    Implementation is based on the work of:
%
%    R.Horaud, F.Forbes, M.Yguel, G.Dewaele and m.Zhang,
%    "Rigid and Articulated point registration with expectation conditional
%    maximization" IEEE Transactions on Pattern Analysis and Machine Intel-
%    ligence, vol. 33, no. 3, pp. 587-602, 2011.
%
% version 24 4 2015 14:16 PM

% compute [m x n] matrix of squared Euclidean distances between Y and X
sqe = @(Y,X) sum(bsxfun(@minus,permute(Y,[2 3 1]),permute(X,[3 2 1])).^2,3);

% get sizes
[~,m] = size(Y); [~,n] = size(X);

% {maxNumIter x 2} x [3 x 3] , [3 x 1] cell array Thistory
if nargout > 4
    Thistory = cell(maxNumIter,2);
end

% [3 x 3] initialise R
R = eye(3);

% [3 x 1] initialise t by the arithmetic mean of Y
t = sum(Y,2) / m;

% [n x 1] initialise variances by the max distance of any points in Y,T(X)
v = 1e3 * ones(n,1);

for iter = 1:maxNumIter
    
    % E step
    
    % [m x n] matrix of Eucledean distances || y_j - Rx_i + t ||_2^2
    a = sqe(Y,bsxfun(@plus,R*X,t));
    
    % [m x n] unnormalised matrix of posteriors  v^-3/2*exp(-.5/v*||.||)
    a = bsxfun(@times, transpose( v.^-1.5 ), exp(bsxfun(@rdivide, a, -2*transpose(v) )));
    
    % [m x n] normalise a by adding an outlier summand
    a = bsxfun(@rdivide,a,sum(a,2) + 1e-30);
    
    % M step 
    
    % [3 x n] weighted pseudo observation
    W = bsxfun( @rdivide, Y*a, transpose(v) );
    
    % [n x 1] weights
    b = transpose( sum(a,1) ) ./ v;
    
    % [3 x 1] mean of X
    mX = X*b;
    
    % [3 x 1] mean of W
    mW = sum(W,2);
    
    % [3 x 3] SVD on the crosscorrelation matrix
    [U,~,V] = svd( W*X' - mW*mX'/sum(b) );
    
    % [3 x 3] optimal Rotation
    R = U*diag( [1 1 det(U*V)] )*V';
    
    % [3 x 1] optimal translation
    t = mW/sum(b) - R*mX/sum(b);
    
    % M step
    
    % [n x 1] isotropic variances
    v = transpose(  sum( a .* sqe(Y, bsxfun(@plus,R*X,t)) ) ./ (3*sum(a))  + 1e-7 );
    
    % if requested populate Thistory
    if nargout > 4
        Thistory(iter,:) = { R , t };
    end
    
    fprintf('iter %d of %d\n',iter,maxNumIter);
end


















